namespace NXyn.EmitFSharp
    open System
    open System.Collections.Generic
    open System.Reflection
    open Microsoft.FSharp.Reflection

    module Common =
        let memberInfo_is_Property(x:System.Reflection.MemberInfo) = x.MemberType = System.Reflection.MemberTypes.Property

        let is_FSharpList(t:System.Type)   = t.IsGenericType && t.GetGenericTypeDefinition() = typedefof<list<_>>
        let is_FSharpOption(t:System.Type) = t.IsGenericType && t.GetGenericTypeDefinition() = typedefof<option<_>>
        let is_FSharpEvent(t:System.Type)  = t.IsGenericType && t.GetGenericTypeDefinition() = typedefof<Event<_>>

        let interface_of_Type (s:string) (t:System.Type) = t.GetInterface(s)

        let is_IEnumerable(t:System.Type)  = t.GetInterface("IEnumerable`1") <> null
        let is_ICollection(t:System.Type)  = t.GetInterface("ICollection`1") <> null

        let is_Struct(t:System.Type)       = t.IsSubclassOf(typeof<System.ValueType>)
        
        let is_FSharpTuple(t:System.Type)  = Microsoft.FSharp.Reflection.FSharpType.IsTuple(t)
        let is_FSharpUnion(t:System.Type)  = Microsoft.FSharp.Reflection.FSharpType.IsUnion(t)
        let is_FSharpRecord(t:System.Type) = Microsoft.FSharp.Reflection.FSharpType.IsRecord(t)

        let tupleElements_of_Type(t:System.Type) = Microsoft.FSharp.Reflection.FSharpType.GetTupleElements(t)
        let unionCases_of_Type(t:System.Type)    = Microsoft.FSharp.Reflection.FSharpType.GetUnionCases(t)

        let propertyType_of_UnionCaseInfo(uc:Microsoft.FSharp.Reflection.UnionCaseInfo) =
            uc.GetFields() |> Array.map (fun (x:System.Reflection.PropertyInfo) -> x.PropertyType)

        let elementType_of_Type(t:System.Type) = t.GetElementType()

        let firstGenericArgument_of_Type(t:System.Type) = (t.GetGenericArguments()).[0]
        let genericArguments_of_Type(t:System.Type) = t.GetGenericArguments()

        let propertyMembers_of_Type(t:System.Type) = t.GetMembers() |> Array.filter memberInfo_is_Property


        let builtinFSharpTypes = 
            [|
             typeof<bigint>,     "bigint";
             typeof<bool>,       "bool";
             typeof<byte>,       "byte";
             typeof<char>,       "char";
             typeof<decimal>,    "decimal";
             typeof<double>,     "double";
             typeof<int16>,      "int16";
             typeof<int32>,      "int32";
             typeof<int64>,      "int64";
             typeof<nativeint>,  "nativeint";
             typeof<sbyte>,      "sbyte";
             typeof<single>,     "single";
             typeof<string>,     "string";
             typeof<uint16>,     "uint16";
             typeof<uint32>,     "uint32";
             typeof<uint64>,     "uint64";
             typeof<unativeint>, "unativeint";
             typeof<unit>,       "unit";
             |]

        let builtinFSharpTypes_map = dict builtinFSharpTypes

        let tryFindName_of_BuiltinFSharpType (t:System.Type) =
            match builtinFSharpTypes_map.TryGetValue(t) with
                | (true, x) -> Some(x)
                | (false,_) -> None

        let isBuiltinFSharpType(t:System.Type) = builtinFSharpTypes_map.ContainsKey(t)



        let BuiltinFSharpTypeNames = (* quasi-DUPLICATES above! *)
            [
                typeof<sbyte>, "sbyte"; typeof<byte>,   "byte";
                typeof<int16>, "int16"; typeof<uint16>, "uint16";
                typeof<int32>, "int32"; typeof<uint32>, "uint32";
                typeof<int64>, "int64"; typeof<uint64>, "uint64";
                typeof<bigint>,"bigint";
                typeof<char>, "char";
                typeof<string>, "string";
                typeof<single>, "single"; typeof<double>, "double";
                typeof<nativeint>, "nativeint"; typeof<unativeint>, "unativeint";
                typeof<bool>, "bool";
                typeof<unit>, "unit";
            ]


        (*
             let IsFlat(t:Type) =
        if List.exists (fun (t',_) -> t=t') BuiltinFSharpTypeNames then true
        elif FSharpType.IsUnion t then Array.forall (fun (uc:UnionCaseInfo) -> uc.GetFields() = [||]) (FSharpType.GetUnionCases t)
        else false
        *)

        let ClippedFSharpTypeName (s:string) =
            let x = let n = s.IndexOf('`') in if n < 0 then s else s.Substring(0,n)
            if x = "FSharpOption" then "Option"
            else x


        let BuiltinNamesMap =
            let m = new Dictionary<Type,string> ()
            for (t,name) in BuiltinFSharpTypeNames do m.Add (t,name)
            m

        let FSharpTypeName (t:Type) =
            let rec AppendFSharpUnionType (sb:Text.StringBuilder) (t:Type) =
                if BuiltinNamesMap.ContainsKey(t) then sb.Append BuiltinNamesMap.[t]
                else if not t.IsGenericType then sb.Append t.Name
                else
                    let sb = sb.Append (ClippedFSharpTypeName t.Name)
                    let mutable sb = sb.Append '<'
                    let t_args = t.GetGenericArguments ()
                    for i = 0 to t_args.Length - 2 do sb <- (AppendFSharpUnionType sb t_args.[i]).Append ','
                    let sb = AppendFSharpUnionType sb t_args.[t_args.Length - 1]
                    sb.Append '>'
            (AppendFSharpUnionType (new Text.StringBuilder()) t).ToString()

        (* does not include generic type parameters *)
        let FSharpTypeName' (t:Type) = if BuiltinNamesMap.ContainsKey(t) then BuiltinNamesMap.[t]
                                                                         else ClippedFSharpTypeName(t.Name)

        let CollectTypes (t:Type) : list<Type> =
            let set = new HashSet<Type>()

            let rec loopT (t:PropertyInfo) = loop t.PropertyType
            and loop (t:Type) =
                if   not (set.Add(t))      then ()
                elif t.IsValueType          then ()
                else
                    match t.GetInterfaces() with
                        | null -> ()
                        | ifaces ->
                            for iface in ifaces do
                                match iface.GetGenericArguments() with
                                    | null -> ()
                                    | args -> for x in args do loop x

                    if   FSharpType.IsUnion t  then for x in FSharpType.GetUnionCases   t do for y in x.GetFields () do loopT y (* in for i = 0 to Array.length xs - 1 do Array.iter loopT (xs.[i].GetFields ()) *)
                    elif FSharpType.IsRecord t then for x in FSharpType.GetRecordFields t do loopT x (* ENUM *)
                    elif t.HasElementType      then loop (t.GetElementType ())
                    elif t.IsGenericType       then for x in t.GetGenericArguments () do loop x

            loop t
            let acc = ref []
            for x in set do acc := x :: !acc
            !acc

        let GetRelevantTypes allTypes (mtype:Type) =
            let newTypes = List.filter (fun (k:Type) -> (List.exists ((=)mtype) (CollectTypes k))) allTypes
            newTypes

        let WriteList (w:IO.TextWriter) (sStart:string) (sSeparator:string) (sEnd:string) (n:int) (f:int -> string) =
            w.Write(sStart)
            for i = 0 to n - 2 do w.Write(f i + sSeparator)
            w.Write(f (n-1) + sEnd)

        let WriteListLine (w:IO.TextWriter) (sStart:string) (sSeparator:string) (sEnd:string) (n:int) (f:int -> string) =
            w.Write(sStart)
            for i = 0 to n - 2 do w.Write(f i + sSeparator)
            w.WriteLine(f (n-1) + sEnd)

        let UnionCaseInfo_equal (uc1:UnionCaseInfo) (uc2:UnionCaseInfo) =
            uc1.Name = uc2.Name
            && (let (s1,s2) = (uc1.GetFields (), uc2.GetFields ()) in
                s1.Length = s2.Length
                && Array.forall2 (fun (x1:PropertyInfo) (x2:PropertyInfo) -> x1.PropertyType = x2.PropertyType) s1 s2)


        let fullName_of_type(t:System.Type) =
            let name = ClippedFSharpTypeName(t.Name)
            match (t.DeclaringType,t.Namespace) with
                | (null,(null|"")) -> name
                | (dtyp,nameSpace) ->
                    match sprintf "%s.%s" (if dtyp <> null then dtyp.FullName else nameSpace) name with
                        | "Microsoft.FSharp.Collections.FSharpMap" -> "Microsoft.FSharp.Collections.Map"
                        | "Microsoft.FSharp.Collections.FSharpSet" -> "Microsoft.FSharp.Collections.Set"
                        | s -> s


        (* special fsharp types! *)
        let fsharpTypeName_of_Type =
            let rec loop1(t:System.Type) =
                let (found,name) = t |> builtinFSharpTypes_map.TryGetValue
                if   found              then name
                elif t.IsArray          then sprintf "%s [%s]"    (t |> elementType_of_Type |> loop1) (String.replicate (t.GetArrayRank() - 1) ",")
                elif is_FSharpOption(t) then sprintf "option<%s>" (t |> firstGenericArgument_of_Type |> loop1)
                elif is_FSharpList(t)   then sprintf "list<%s>"   (t |> firstGenericArgument_of_Type |> loop1)
                elif is_FSharpTuple(t)  then sprintf "(%s)"       (t |> tupleElements_of_Type |> NXyn.Aux.IOMisc.StringMany "" " * " "" loop1)
                elif t.IsGenericType    then sprintf "%s<%s>"     (t |> fullName_of_type) (t |> genericArguments_of_Type |> NXyn.Aux.IOMisc.StringMany "" "," "" loop1) 
                                        else t |> fullName_of_type
            fun(t:System.Type) -> loop1 t


        let fsharpTypeName_of_Type' =
            let rec loop1(t:System.Type) =
                let (found,name) = t |> builtinFSharpTypes_map.TryGetValue
                if   found              then name
                elif t.IsArray          then sprintf "%s [%s]"    (t |> elementType_of_Type |> loop1) (String.replicate (t.GetArrayRank() - 1) ",") (* ??? *)
                elif is_FSharpOption(t) then sprintf "option<%s>" (t |> firstGenericArgument_of_Type |> loop1) (* ??? *)
                elif is_FSharpList(t)   then sprintf "list<%s>"   (t |> firstGenericArgument_of_Type |> loop1) (* ??? *)
                elif is_FSharpTuple(t)  then sprintf "(%s)"       (t |> tupleElements_of_Type |> NXyn.Aux.IOMisc.StringMany "" " * " "" loop1)
                elif t.IsGenericType    then t |> fullName_of_type |> ClippedFSharpTypeName
                                        else t |> fullName_of_type
            fun(t:System.Type) -> loop1 t


        let CollectTypes2(t:System.Type) : array<System.Type> =
            let set = new System.Collections.Generic.HashSet<System.Type>()

            let rec loopT (t:PropertyInfo) = loop t.PropertyType
            and loop (t:System.Type) =
                if   not (set.Add(t))      then ()
                elif t.IsValueType         then ()     (* struct is also value-type! *)
                else
                    match t.GetInterfaces() with
                        | null -> ()
                        | ifaces -> 
                        for iface in ifaces do 
                        match iface.GetGenericArguments() with
                            | null -> ()
                            | args -> for x in args do loop x

                    if FSharpType.IsUnion t    then for x in FSharpType.GetUnionCases   t do for y in x.GetFields () do loopT y (* in for i = 0 to Array.length xs - 1 do Array.iter loopT (xs.[i].GetFields ()) *)
                    elif FSharpType.IsRecord t then for x in FSharpType.GetRecordFields t do loopT x
                    elif t.HasElementType      then loop (t.GetElementType ())
                    elif t.IsGenericType       then for x in t.GetGenericArguments () do loop x
 
            loop t
            let acc = Array.zeroCreate set.Count
            set.CopyTo(acc,0)
            acc



        (* somewhat imprecise ?? *)
(*        let RemoveIdenticalTypes (mtype:Type) (userFilter:PropertyInfo -> bool) newTypes =
            newTypes
*)
        let RemoveIdenticalTypes (mtype:Type) (userFilter:PropertyInfo -> bool) newTypes =
            let rec loop (newTypes:Map<string,Type>) =

(* assert mtype <> t *)
                let isIdentity_of_type (t:Type) =
                    let check1 (f:PropertyInfo) =        (* explain that test, and why is it both sufficient and necessary *)
                        let t' = f.PropertyType
                        if t' = t then true 
                        else if t' = mtype && userFilter f then false
                        else if t'.FullName <> t.FullName && newTypes.ContainsKey(t'.FullName) then false
                        (* else if newTypes.ContainsKey(t'.FullName) then false *)
                        else true

                    if FSharpType.IsRecord t then
                        FSharpType.GetRecordFields(t)
                        |> Array.forall check1
                    elif FSharpType.IsUnion t then

                        FSharpType.GetUnionCases(t)
                        |> Array.forall (fun (uc:UnionCaseInfo) -> uc.GetFields() |> Array.forall check1)
  
                    else (* what bout: array, list, option, icollection, *)
                        (* boiler *)
                        let t' = t
                        if t'.FullName <> t.FullName && newTypes.ContainsKey(t'.FullName) then false
                        (* else if newTypes.ContainsKey(t'.FullName) then false *)
                        else true

                        (* true *) (* failwith (Printf.sprintf "EmitFSharpCommon.RemoveIdenticalTypes: not_implemented %A" t) *)

                let newTypes1 = Map.filter (fun _ t -> not(isIdentity_of_type(t))) newTypes
                
(*                let newTypes1 = Map.filter (fun _ t -> t = mtype || not(isIdentity_of_type(t))) newTypes *)
                if newTypes1.Count < newTypes.Count then
                    loop newTypes1
                else
                    List.map snd (Map.toList(newTypes))
            loop(Map.ofList(List.map (fun(t:Type) -> (t.FullName,t)) newTypes))

