module GA

open System
open System.Collections.Generic

(***)

type SelectionMethod = TournamentSelection of float * int
                     | RouletteWheelSelection
                     | RandomSelection

type Population<'i when 'i : equality> = (float * 'i) array

let array_of_collection(c:ICollection<'a>) = let xs = Array.zeroCreate c.Count in c.CopyTo(xs,0); xs

let binSearchProximate (arr:'a []) (el:'a) = let p = Array.BinarySearch(arr,el,ComparisonIdentity.FromFunction(compare))
                                             if p < 0 then ~~~(p) else p

type t<'i when 'i : equality and 'i :> System.ICloneable>(Evaluate            : 'i -> float,
                                                          Crossover           : 'i -> 'i -> 'i [],
                                                          Mutate              : 'i -> unit,
                                                          Generate            : unit -> 'i,
                                                          ?Verbosity          : Int32,
                                                          ?QuitIfNoNewBestFor : Int32,
                                                          ?SoftTimeLimit      : TimeSpan,
                                                          ?MutateProbability  : Double, 
                                                          ?PopulationSize     : Int32,
                                                          ?RNG                : Random,
                                                          ?Selection          : SelectionMethod) =

     let verbosity           = defaultArg Verbosity           0
     let mutateProbability   = defaultArg MutateProbability   0.0001
     let populationSize      = defaultArg PopulationSize      1000
     let selection           = defaultArg Selection           (TournamentSelection(0.8,2))
     let softTimeLimit       = defaultArg SoftTimeLimit       TimeSpan.MaxValue
     let quitIfBestUnchFor   = defaultArg QuitIfNoNewBestFor  20
     let rng                 = match RNG with Some(r) -> r | None -> new System.Random()

     let tagWithFitness (items:'i []) =
                        items 
                        |> Array.map (fun(x) -> async { let f = Evaluate(x) in return((f,x)) })
                        |> Async.Parallel 
                        |> Async.RunSynchronously
     
     member x.Next (pop:Population<'i>) =
                   let n     = pop.Length
                   let pick1 = match selection
                                with | RandomSelection -> fun () -> rng.Next(n)
                                     | RouletteWheelSelection -> let ks = Array.create n 0.0
                                                                 let (sum,_) = Array.fold (fun (f_sum,i) (f,_) -> ks.[i] <- f_sum; (f_sum + f, i+1)) (0.0,0) pop
                                                                 fun () -> min (n-1) (binSearchProximate ks (sum * rng.NextDouble()))
                                     | TournamentSelection(p,k) -> let (p0,p1) = (ref 0.0, ref p)
                                                                   let pvals = Array.init k (fun i -> p0 := !p0 + !p1; p1 := !p1 * (1.0 - p); !p0)
                                                                   pvals.[k-1] <- 1.0

                                                                   let tour = Array.zeroCreate(k)
                                                                   fun () -> for i = 0 to k - 1 do let ix = rng.Next(n) in tour.[i] <- (fst(pop.[ix]),ix)
                                                                             Array.sortInPlaceWith (fun (k1,_) (k2,_) -> compare k2 k1) tour (* put best entry first *)
                                                                             let ix = binSearchProximate pvals (rng.NextDouble())
                                                                             if ix >= tour.Length then failwithf "GA.Partition: ix out-of-bound; pvals=%A, tour=%A" pvals tour
                                                                             tour.[ix] |> snd

                   let pairs = new ResizeArray<'i>()
                   for i = 1 to min n (populationSize / 2) do
                         let i1 = pick1()
                         let i2 = pick1()
                         let ((_,xx),(_,yy)) = (pop.[i1],pop.[i2])
                         if i1 = i2 then let xz = xx.Clone() :?> 'i
                                         Mutate(xz)
                                         pairs.Add(xz)
                                    else Crossover xx yy |> Array.iter (fun(xz) -> if rng.NextDouble() < mutateProbability then Mutate(xz)
                                                                                   pairs.Add(xz))
                   pairs.ToArray() |> tagWithFitness

  
     member x.Do (maxNgen:Int32, population0:'i [], endSelSize:Int32) =
                 let t0 = DateTime.Now
                 if population0 = [||] then failwith "GA.Do: population0 = [||]"

                 (* initialize mutable state *)
                 let mutable n_generations                          = 0
                 let mutable population                             = population0 |> tagWithFitness
                 let mutable (best_since,(best_fitness,best_genes)) = (0,population |> Array.maxBy fst)

                 (* GA main loop *)
                 while maxNgen           > n_generations                                      (* while the max # of generations has not passed ... *)
                    && softTimeLimit     > DateTime.Now - t0                                  (*   and run time has not run exceeded the given limit ... *)
                    && quitIfBestUnchFor > n_generations - best_since                         (*   and best is `younger' than `quitIfBestUnchFor' generations ... *)
                       do (* advance number of generations by one *)
                          n_generations <- n_generations + 1

                          (* variation proposes, selection disposes ... *)
                          population <- x.Next(population)

                          if population <> [||] then (* new best? if so, update *)
                                                     let (fitnessB,genesB) = population |> Array.maxBy fst
                                                     if fitnessB > best_fitness then best_since <- n_generations
                                                                                     best_genes <- genesB
                                                                                     best_fitness <- fitnessB
                                                     (* dump auxiliary information *)
                                                     
                                                     if verbosity > 0 then let fitnesses = population |> Array.map fst
                                                                           eprintfn "[Population(%d): size=%d, minF=%f, avgF=%f maxF=%f]" n_generations
                                                                                                                                          population.Length
                                                                                                                                          (fitnesses |> Array.min)
                                                                                                                                          (fitnesses |> Array.average)
                                                                                                                                          (fitnesses |> Array.max)
                                                                                                                                  
                 (* sort population entries by decreasing fitness, return `endSelSize' best *)
                 let tmp = Array.append [|(best_fitness,best_genes)|] population
                 Array.sortInPlaceBy (fst >> (~-)) tmp
                 Array.sub tmp 0 (min (Array.length tmp) endSelSize)
                

