module Generate_InstrIO

open System
open System.Collections.Generic

open System.Reflection
open Microsoft.FSharp.Reflection

open X64instr

open NXyn.EmitFSharp
open NXyn.EmitFSharp.Common

open NXyn.Aux.IOMisc

open NXyn.Aux

let opTypes = FSharpType.GetUnionCases(typeof<Instr>)
              |> Array.map (fun(uc:UnionCaseInfo) -> let fs = uc.GetFields()
                                                     (uc.Name, fs.[0].PropertyType, if fs.Length > 1 then Some(fs.[1].PropertyType) else None))

(***)

(* types supported as arguments of `Instr.Op' types *)
let instrOpIO(t:Type) = if   t = typeof<Lit>  then ("Match_Lit",  "Item_of_Lit")
                        elif t = typeof<M>    then ("Match_M",    "Item_of_M")
                        elif t = typeof<NM64> then ("Match_NM64", "Item_of_NM64")
                        elif t = typeof<NR64> then ("Match_NR64", "Item_of_NR64")
                        else failwithf "GenerateModule_InstrIO.instrOpIO: unsupported sub-type `%A'" t

let inputInstrOpArg(t:Type)  = t |> instrOpIO |> fst
let outputInstrOpArg(t:Type) = t |> instrOpIO |> snd

(***)

(* map record field to assembly argument position (AT&T) *)
let AttPos_of_PropertyInfo (field:PropertyInfo) =
                           match field.GetCustomAttributes(typeof<AttPosAttribute>,true)
                            with | [| n |] -> (n :?> AttPosAttribute).GetValue
                                 | _ -> failwith "GenerateModule_InstrIO.AttPos_of_PropertyInfo"

(***)

(* parser *)
let emit___Instr_of_ELine(w:IO.TextWriter) =
    (* build table *)
    let dict = new Dictionary<string, string list>()
    opTypes
    |> Array.iter (fun (name,ty,argsoption) ->
                       let args = match argsoption
                                   with | None -> [||]
                                        | Some(x) -> FSharpType.GetRecordFields(x)
                                                     |> Array.map (fun(f) -> (AttPos_of_PropertyInfo(f),(f.PropertyType.Name,f.Name)))
                                                     |> Array.sortBy fst

                       FSharpType.GetUnionCases(ty)
                       |> Array.iter (fun uc ->
                                          let fields = uc.GetFields()

                                          uc.GetCustomAttributes (typeof<AKAAttribute>)
                                          |> Array.map (fun(a) -> string_of_AKA(a) |> fun (s:String) -> s.Split([| ',' |], StringSplitOptions.RemoveEmptyEntries))
                                          |> Array.concat
                                          |> Array.append [|uc.Name|]
                                          |> Array.iter (fun n ->
                                                             let w = new IO.StringWriter()
                                                             fprintf w "[%t%t] -> Some(Instr.%s(%s.%s%t%t))"
                                                                         (fields |> WriteManyI_to "" ";" ";" (fun i (f:PropertyInfo) w -> fprintf w "%s l%d" (inputInstrOpArg(f.PropertyType)) i))
                                                                           (args |> WriteMany_to  "" ";" "" (fun (key,(matchName,_)) w -> fprintf w "Match_%s a%d" matchName key))
                                                                                             name
                                                                                                ty.Name
                                                                                                   uc.Name
                                                                                                     (fields |> WriteManyI_to "(" "," ")" (fun i _ w -> fprintf w "l%d" i))
                                                                                                       (args |> WriteMany_to ",{" ";" "}" (fun (key,(_,fieldName)) w ->
                                                                                                                                               fprintf w "%s = a%d" fieldName key))
                                                             let n = (n:String).ToLower()
                                                             let (found,lines) = dict.TryGetValue(n)
                                                             dict.[n] <-  (w.ToString()) :: (if found then lines else []))))

    let all = dict |> ArrayMisc.array_of_collection |> Array.map (function KeyValue(mnemonic,lines) -> (mnemonic,lines))

    (* emit lookup and dispatch code *)
    fprintfn w "let Instr_of_ELine ="
    all |> Array.iter (fun (mnemonic,lines) ->
                           fprintfn w "  let H_%s = function" mnemonic
                           lines |> List.iter (fun line -> fprintfn w "    | %s" line)
                           fprintfn w "    | _ -> None")
    fprintfn w "  let Hdict = new System.Collections.Generic.Dictionary<string,(Ast.Item list -> X64instr.Instr option)>(%d)" dict.Count
    fprintfn w "  let Hinit ="
    all |> Array.iter (fun (mnemonic,_) -> fprintfn w "    Hdict.Add(\"%s\", H_%s)" mnemonic mnemonic)
    fprintfn w "  fun(mnemonic,args) -> let(found,f) = Hdict.TryGetValue(mnemonic) in if found then f(args) else None"

(***)

(* unparser *)
let emit___TagString_of_ (t:Type) (name:string) (w:IO.TextWriter) = 
    if t.ContainsGenericParameters then failwith "EmitFSharpDispatchUnion.OutputCode: open types not supported"
    let tName = ClippedFSharpTypeName(t.Name)
    fprintfn w "let TagString_of_%s = function" name
    FSharpType.GetUnionCases(t)
    |> Array.iter (fun uc ->
                       let fs = uc.GetFields()
                       fprintfn w "  | %s.%s%s -> (\"%s\",[%t])"
                                       tName
                                          uc.Name
                                            (string_aI "(" ")" fs.Length)
                                                     (uc.Name.Replace("'","").ToLower()) (* remove ticks "'" from mnemonic names *)
                                                           (WriteManyI_to "" ";" "" (fun i (f:PropertyInfo) w -> fprintf w "%s a%d" (outputInstrOpArg f.PropertyType) i) fs))


let emit___ELine_of_Instr(w:IO.TextWriter) =
    w.WriteLine("let ELine_of_Instr = function")
    opTypes
    |> Array.iter (fun (name,_,argoption) ->
                       fprintf w "  | Instr.%s" name
                       match argoption
                        with | None    -> fprintfn w "(op) -> let (s,ls) = TagString_of_%s op in Tagged(s,ls)" name
                             | Some(a) -> fprintfn w "(op,args) -> let ls1 = [%t] in let (s,ls0) = TagString_of_%s op in Tagged(s, ls0 @ ls1)"
                                                                              (FSharpType.GetRecordFields(a)
                                                                               |> Array.sortBy AttPos_of_PropertyInfo
                                                                               |> WriteMany_to "" ";" "" (fun f w -> fprintf w "Item_of_%s args.%s" f.PropertyType.Name f.Name))
                                                                                                                name)
(***)

let k_revCons = Some "revCons"
let k_nil     = Some "[]"
let k_lOr     = Some "||"
let k_bOr     = Some "|||"
let k_enum0   = Some "(enum 0)"

let main() =
    let w = new IO.StreamWriter("InstrIO.fs")

    let Instr_DepsOn = CollectTypes typeof<Instr>
    let M_DepsOn     = CollectTypes typeof<M>
    let Iexpr_DepsOn = CollectTypes typeof<Iexpr>

    let Instr_X_RelevantTypes     = GetRelevantTypes Instr_DepsOn typeof<X>
    let Instr_Y_RelevantTypes     = GetRelevantTypes Instr_DepsOn typeof<Y>
    let Instr_Id_RelevantTypes    = GetRelevantTypes Instr_DepsOn typeof<Id>
    let Instr_R_RelevantTypes     = GetRelevantTypes Instr_DepsOn typeof<R>
    let Instr_R8_RelevantTypes    = GetRelevantTypes Instr_DepsOn typeof<R8>
    let Instr_R16_RelevantTypes   = GetRelevantTypes Instr_DepsOn typeof<R16>
    let Instr_R32_RelevantTypes   = GetRelevantTypes Instr_DepsOn typeof<R32>
    let Instr_R64_RelevantTypes   = GetRelevantTypes Instr_DepsOn typeof<R64>
    let Instr_M_RelevantTypes     = GetRelevantTypes Instr_DepsOn typeof<M>
    let Instr_M8_RelevantTypes    = GetRelevantTypes Instr_DepsOn typeof<M8>
    let Instr_M16_RelevantTypes   = GetRelevantTypes Instr_DepsOn typeof<M16>
    let Instr_M32_RelevantTypes   = GetRelevantTypes Instr_DepsOn typeof<M32>
    let Instr_M64_RelevantTypes   = GetRelevantTypes Instr_DepsOn typeof<M64>
    let Instr_M128_RelevantTypes  = GetRelevantTypes Instr_DepsOn typeof<M128>
    let Instr_M256_RelevantTypes  = GetRelevantTypes Instr_DepsOn typeof<M256>

    let Iexpr_Id_RelevantTypes    = GetRelevantTypes Iexpr_DepsOn typeof<Id>
    
    let M_R32_RelevantTypes       = GetRelevantTypes M_DepsOn     typeof<R32>
    let M_R64_RelevantTypes       = GetRelevantTypes M_DepsOn     typeof<R64>
    let M_Id_RelevantTypes        = GetRelevantTypes M_DepsOn     typeof<Id>

    (* => Aux.TypeMisc: Type_HasCustomAttribute : IsDefined? *)
    let typeHasCustomAttribute (att:Type) (fs:PropertyInfo) = fs.IsDefined(att,true)
    let typeHasSomeCustomAttribute (atts:Type []) (fs:PropertyInfo) = atts |> Array.exists (fun(att) -> typeHasCustomAttribute att fs)
       
    [
        [|
         WriteString_to "(* This source file was generated automatically by `GenerateModule_InstrIO' --- DO NOT EDIT *)"
         WriteString_to (sprintf "(* creation date %A *)" DateTime.Now)
         WriteNewLine_to
         WriteString_to "module InstrIO"
         WriteNewLine_to
         WriteString_to "open X64instr"
         WriteString_to "open Ast"
         WriteString_to "open NXyn.Aux.ListMisc"
         WriteNewLine_to

         WriteString_to JSONio.ModulePrelude

         JSONio.WriteUnparser "JSONvalue_of_Instr" typeof<Instr>
         JSONio.WriteParser   "Instr_of_JSONvalue" typeof<Instr>

         (* WriteMapper *)
         Map.OutputCode "Map_Xd_in_Instr"              Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (typeHasCustomAttribute typeof<DAttribute>)
         Map.OutputCode "Map_Xds_in_Instr"             Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (typeHasCustomAttribute typeof<DSAttribute>)
         Map.OutputCode "Map_Xs_in_Instr"              Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (typeHasCustomAttribute typeof<SAttribute>)
         Map.OutputCode "Map_Xds_and_Xs_in_Instr"      Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (typeHasSomeCustomAttribute [| typeof<SAttribute>; typeof<DSAttribute> |])
         Map.OutputCode "Map_X_in_Instr"               Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (fun _ -> true)

         Map.OutputCode "Map_Yd_in_Instr"              Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (typeHasCustomAttribute typeof<DAttribute>)
         Map.OutputCode "Map_Yds_in_Instr"             Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (typeHasCustomAttribute typeof<DSAttribute>)
         Map.OutputCode "Map_Ys_in_Instr"              Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (typeHasCustomAttribute typeof<SAttribute>)
         Map.OutputCode "Map_Yds_and_Ys_in_Instr"      Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (typeHasSomeCustomAttribute [| typeof<SAttribute>; typeof<DSAttribute> |])
         Map.OutputCode "Map_Y_in_Instr"               Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (fun _ -> true)

         Map.OutputCode "Map_R64d_in_Instr"            Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (typeHasCustomAttribute typeof<DAttribute>)
         Map.OutputCode "Map_R64ds_in_Instr"           Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (typeHasCustomAttribute typeof<DSAttribute>)
         Map.OutputCode "Map_R64s_in_Instr"            Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (typeHasCustomAttribute typeof<SAttribute>)
         Map.OutputCode "Map_R64ds_and_R64s_in_Instr"  Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (typeHasSomeCustomAttribute [| typeof<SAttribute>; typeof<DSAttribute> |])
         Map.OutputCode "Map_R64_in_Instr"             Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (fun _ -> true)

         Map.OutputCode "Map_R32d_in_Instr"            Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (typeHasCustomAttribute typeof<DAttribute>)
         Map.OutputCode "Map_R32ds_in_Instr"           Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (typeHasCustomAttribute typeof<DSAttribute>)
         Map.OutputCode "Map_R32s_in_Instr"            Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (typeHasCustomAttribute typeof<SAttribute>)
         Map.OutputCode "Map_R32ds_and_R32s_in_Instr"  Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (typeHasSomeCustomAttribute [| typeof<SAttribute>; typeof<DSAttribute> |])
         Map.OutputCode "Map_R32_in_Instr"             Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (fun _ -> true)

         Map.OutputCode "Map_R16d_in_Instr"            Instr_R16_RelevantTypes typeof<Instr> typeof<R16> (typeHasCustomAttribute typeof<DAttribute>)
         Map.OutputCode "Map_R16ds_in_Instr"           Instr_R16_RelevantTypes typeof<Instr> typeof<R16> (typeHasCustomAttribute typeof<DSAttribute>)
         Map.OutputCode "Map_R16s_in_Instr"            Instr_R16_RelevantTypes typeof<Instr> typeof<R16> (typeHasCustomAttribute typeof<SAttribute>)
         Map.OutputCode "Map_R16ds_and_R16s_in_Instr"  Instr_R16_RelevantTypes typeof<Instr> typeof<R16> (typeHasSomeCustomAttribute [| typeof<SAttribute>; typeof<DSAttribute> |])

         Map.OutputCode "Map_R8d_in_Instr"             Instr_R8_RelevantTypes  typeof<Instr> typeof<R8> (typeHasCustomAttribute typeof<DAttribute>)
         Map.OutputCode "Map_R8ds_in_Instr"            Instr_R8_RelevantTypes  typeof<Instr> typeof<R8> (typeHasCustomAttribute typeof<DSAttribute>)
         Map.OutputCode "Map_R8s_in_Instr"             Instr_R8_RelevantTypes  typeof<Instr> typeof<R8> (typeHasCustomAttribute typeof<SAttribute>)
         Map.OutputCode "Map_R8ds_and_R8s_in_Instr"    Instr_R8_RelevantTypes  typeof<Instr> typeof<R8> (typeHasSomeCustomAttribute [| typeof<SAttribute>; typeof<DSAttribute> |])

         Map.OutputCode "Map_R_in_Instr"               Instr_R_RelevantTypes   typeof<Instr> typeof<R>  (fun _ -> true)

         Map.OutputCode "Map_M_in_Instr"               Instr_M_RelevantTypes   typeof<Instr> typeof<M>  (fun _ -> true)

         Map.OutputCode "Map_Id_in_Instr"              Instr_Id_RelevantTypes  typeof<Instr> typeof<Id> (fun _ -> true)

         (* WriteFolder *)
         Fold.OutputCode "Ys_of_Instr"       Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil 
         Fold.OutputCode "Yds_of_Instr"      Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "Yd_of_Instr"       Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil
         Fold.OutputCode "Y_of_Instr"        Instr_Y_RelevantTypes   typeof<Instr> typeof<Y>   (fun _ -> true)                              k_revCons k_nil

         Fold.OutputCode "Xs_of_Instr"       Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil 
         Fold.OutputCode "Xds_of_Instr"      Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "Xd_of_Instr"       Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil
         Fold.OutputCode "X_of_Instr"        Instr_X_RelevantTypes   typeof<Instr> typeof<X>   (fun _ -> true)                              k_revCons k_nil

         Fold.OutputCode "R8s_of_Instr"      Instr_R8_RelevantTypes  typeof<Instr> typeof<R8>  (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "R8ds_of_Instr"     Instr_R8_RelevantTypes  typeof<Instr> typeof<R8>  (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "R8d_of_Instr"      Instr_R8_RelevantTypes  typeof<Instr> typeof<R8>  (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil
         Fold.OutputCode "R8_of_Instr"       Instr_R8_RelevantTypes  typeof<Instr> typeof<R8>  (fun _ -> true)                              k_revCons k_nil

         Fold.OutputCode "R16s_of_Instr"     Instr_R16_RelevantTypes typeof<Instr> typeof<R16> (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "R16ds_of_Instr"    Instr_R16_RelevantTypes typeof<Instr> typeof<R16> (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "R16d_of_Instr"     Instr_R16_RelevantTypes typeof<Instr> typeof<R16> (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil
         Fold.OutputCode "R16_of_Instr"      Instr_R16_RelevantTypes typeof<Instr> typeof<R16> (fun _ -> true)                              k_revCons k_nil

         Fold.OutputCode "R32s_of_Instr"     Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "R32ds_of_Instr"    Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "R32d_of_Instr"     Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil
         Fold.OutputCode "R32_of_Instr"      Instr_R32_RelevantTypes typeof<Instr> typeof<R32> (fun _ -> true)                              k_revCons k_nil

         Fold.OutputCode "R64s_of_Instr"     Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "R64ds_of_Instr"    Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "R64d_of_Instr"     Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil
         Fold.OutputCode "R64_of_Instr"      Instr_R64_RelevantTypes typeof<Instr> typeof<R64> (fun _ -> true)                              k_revCons k_nil

         Fold.OutputCode "R32_of_M"          M_R32_RelevantTypes     typeof<M>     typeof<R32> (fun _ -> true)                              k_revCons k_nil
         Fold.OutputCode "R64_of_M"          M_R64_RelevantTypes     typeof<M>     typeof<R64> (fun _ -> true)                              k_revCons k_nil

         WriteString_to "let R_of_Instr(i:Instr)   = List.map Rbyte (R8_of_Instr(i))   @ List.map Rword (R16_of_Instr(i))   @ List.map Rlong (R32_of_Instr(i))   @ List.map Rquad (R64_of_Instr(i))"
         WriteString_to "let Rs_of_Instr(i:Instr)  = List.map Rbyte (R8s_of_Instr(i))  @ List.map Rword (R16s_of_Instr(i))  @ List.map Rlong (R32s_of_Instr(i))  @ List.map Rquad (R64s_of_Instr(i))"
         WriteString_to "let Rds_of_Instr(i:Instr) = List.map Rbyte (R8ds_of_Instr(i)) @ List.map Rword (R16ds_of_Instr(i)) @ List.map Rlong (R32ds_of_Instr(i)) @ List.map Rquad (R64ds_of_Instr(i))"
         WriteString_to "let Rd_of_Instr(i:Instr)  = List.map Rbyte (R8d_of_Instr(i))  @ List.map Rword (R16d_of_Instr(i))  @ List.map Rlong (R32d_of_Instr(i))  @ List.map Rquad (R64d_of_Instr(i))"

         (***)

         Fold.OutputCode "M8s_of_Instr"     Instr_M8_RelevantTypes   typeof<Instr> typeof<M8>   (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "M8ds_of_Instr"    Instr_M8_RelevantTypes   typeof<Instr> typeof<M8>   (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "M8d_of_Instr"     Instr_M8_RelevantTypes   typeof<Instr> typeof<M8>   (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil

         Fold.OutputCode "M16s_of_Instr"    Instr_M16_RelevantTypes  typeof<Instr> typeof<M16>  (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "M16ds_of_Instr"   Instr_M16_RelevantTypes  typeof<Instr> typeof<M16>  (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "M16d_of_Instr"    Instr_M16_RelevantTypes  typeof<Instr> typeof<M16>  (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil

         Fold.OutputCode "M32s_of_Instr"    Instr_M32_RelevantTypes  typeof<Instr> typeof<M32>  (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "M32ds_of_Instr"   Instr_M32_RelevantTypes  typeof<Instr> typeof<M32>  (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "M32d_of_Instr"    Instr_M32_RelevantTypes  typeof<Instr> typeof<M32>  (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil

         Fold.OutputCode "M64s_of_Instr"    Instr_M64_RelevantTypes  typeof<Instr> typeof<M64>  (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "M64ds_of_Instr"   Instr_M64_RelevantTypes  typeof<Instr> typeof<M64>  (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "M64d_of_Instr"    Instr_M64_RelevantTypes  typeof<Instr> typeof<M64>  (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil

         Fold.OutputCode "M128s_of_Instr"   Instr_M128_RelevantTypes typeof<Instr> typeof<M128> (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "M128ds_of_Instr"  Instr_M128_RelevantTypes typeof<Instr> typeof<M128> (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "M128d_of_Instr"   Instr_M128_RelevantTypes typeof<Instr> typeof<M128> (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil

         Fold.OutputCode "M256s_of_Instr"   Instr_M256_RelevantTypes typeof<Instr> typeof<M256> (typeHasCustomAttribute typeof<SAttribute>)  k_revCons k_nil
         Fold.OutputCode "M256ds_of_Instr"  Instr_M256_RelevantTypes typeof<Instr> typeof<M256> (typeHasCustomAttribute typeof<DSAttribute>) k_revCons k_nil
         Fold.OutputCode "M256d_of_Instr"   Instr_M256_RelevantTypes typeof<Instr> typeof<M256> (typeHasCustomAttribute typeof<DAttribute>)  k_revCons k_nil

         Fold.OutputCode "M_of_Instr"       Instr_M_RelevantTypes    typeof<Instr> typeof<M>    (fun _ -> true)                              k_revCons k_nil (*use?*)

         (***)

         Fold.OutputCode "Id_of_Instr"      Instr_Id_RelevantTypes   typeof<Instr> typeof<Id>   (fun _ -> true)                              k_revCons k_nil
         Fold.OutputCode "Id_of_M"          M_Id_RelevantTypes       typeof<M>     typeof<Id>   (fun _ -> true)                              k_revCons k_nil
         Fold.OutputCode "Id_of_Iexpr"      Iexpr_Id_RelevantTypes   typeof<Iexpr> typeof<Id>   (fun _ -> true)                              k_revCons k_nil
         
         (***)
         
         FoldAttributes.OutputCode "FixedRs_of_Instr"    Instr_DepsOn typeof<Instr> typeof<RsAttribute>  (fun x -> (x :?> RsAttribute).GetValue)  "R." k_revCons k_nil
         FoldAttributes.OutputCode "FixedRds_of_Instr"   Instr_DepsOn typeof<Instr> typeof<RdsAttribute> (fun x -> (x :?> RdsAttribute).GetValue) "R." k_revCons k_nil
         FoldAttributes.OutputCode "FixedRd_of_Instr"    Instr_DepsOn typeof<Instr> typeof<RdAttribute>  (fun x -> (x :?> RdAttribute).GetValue)  "R." k_revCons k_nil
         WriteString_to "let FixedR_of_Instr(i:Instr) = FixedRd_of_Instr(i) @ FixedRds_of_Instr(i) @ FixedRs_of_Instr(i)"

         FoldAttributes.OutputCode "FixedXs_of_Instr"    Instr_DepsOn typeof<Instr> typeof<XsAttribute>  (fun x -> (x :?> XsAttribute).GetValue)  "X." k_revCons k_nil
         FoldAttributes.OutputCode "FixedXds_of_Instr"   Instr_DepsOn typeof<Instr> typeof<XdsAttribute> (fun x -> (x :?> XdsAttribute).GetValue) "X." k_revCons k_nil
         FoldAttributes.OutputCode "FixedXd_of_Instr"    Instr_DepsOn typeof<Instr> typeof<XdAttribute>  (fun x -> (x :?> XdAttribute).GetValue)  "X." k_revCons k_nil
         WriteString_to "let FixedX_of_Instr(i:Instr) = FixedXd_of_Instr(i) @ FixedXds_of_Instr(i) @ FixedXs_of_Instr(i)"

         FoldAttributes.OutputCode "FixedYs_of_Instr"    Instr_DepsOn typeof<Instr> typeof<YsAttribute>  (fun x -> (x :?> YsAttribute).GetValue)  "Y." k_revCons k_nil
         FoldAttributes.OutputCode "FixedYds_of_Instr"   Instr_DepsOn typeof<Instr> typeof<YdsAttribute> (fun x -> (x :?> YdsAttribute).GetValue) "Y." k_revCons k_nil
         FoldAttributes.OutputCode "FixedYd_of_Instr"    Instr_DepsOn typeof<Instr> typeof<YdAttribute>  (fun x -> (x :?> YdAttribute).GetValue)  "Y." k_revCons k_nil
         WriteString_to "let FixedY_of_Instr(i:Instr) = FixedYd_of_Instr(i) @ FixedYds_of_Instr(i) @ FixedYs_of_Instr(i)"
         
         FoldAttributes.OutputCode "Fs_of_Instr"         Instr_DepsOn typeof<Instr> typeof<FsAttribute>  (fun x -> (x :?> FsAttribute).GetValue)  "F." k_revCons k_nil
         FoldAttributes.OutputCode "Fd_of_Instr"         Instr_DepsOn typeof<Instr> typeof<FdAttribute>  (fun x -> (x :?> FdAttribute).GetValue)  "F." k_revCons k_nil

         (***)

         emit___Instr_of_ELine
         |]

        [| "Instr_uses_FixedMem",   typeof<MemAttribute>
           "Instr_may_raise_exn",   typeof<ExnAttribute>
           "Instr_is_Prefix",       typeof<PfxAttribute>
           "Instr_is_FunReturn",    typeof<RetAttribute>
           "Instr_is_NoOp",         typeof<NopAttribute>
           "Instr_is_Vec",          typeof<VecAttribute>
           "Instr_is_Jthru",        typeof<JthruAttribute>
           "Instr_is_Jfollow",      typeof<JfollowAttribute>
           "Instr_may_be_repeated", typeof<RepAttribute>
        |] |> Array.map (fun(name,t) -> FoldAttributes.OutputCode name Instr_DepsOn typeof<Instr> t (fun _ -> "true") "" (Some "||") (Some "false"))

        Array.map (fun (name,t,_) -> emit___TagString_of_ t name) opTypes

        [| emit___ELine_of_Instr |]
    ]
    |> Array.concat
    |> Array.map (fun f -> async { return(let w = new IO.StringWriter() in f(w); w.ToString()) })
    |> Async.Parallel 
    |> Async.RunSynchronously
    |> Array.iter (fun s -> w.WriteLine(s))
  
    w.Close()

main()

(* END *)
